/*
THE COMPUTER CODE CONTAINED HEREIN IS THE SOLE PROPERTY OF PARALLAX
SOFTWARE CORPORATION ("PARALLAX").  PARALLAX, IN DISTRIBUTING THE CODE TO
END-USERS, AND SUBJECT TO ALL OF THE TERMS AND CONDITIONS HEREIN, GRANTS A
ROYALTY-FREE, PERPETUAL LICENSE TO SUCH END-USERS FOR USE BY SUCH END-USERS
IN USING, DISPLAYING,  AND CREATING DERIVATIVE WORKS THEREOF, SO LONG AS
SUCH USE, DISPLAY OR CREATION IS FOR NON-COMMERCIAL, ROYALTY OR REVENUE
FREE PURPOSES.  IN NO EVENT SHALL THE END-USER USE THE COMPUTER CODE
CONTAINED HEREIN FOR REVENUE-BEARING PURPOSES.  THE END-USER UNDERSTANDS
AND AGREES TO THE TERMS HEREIN AND ACCEPTS THE SAME BY USE OF THIS FILE.
COPYRIGHT 1993-1999 PARALLAX SOFTWARE CORPORATION.  ALL RIGHTS RESERVED.
*/

#ifndef _RENDERTHREADS_H
#define _RENDERTHREADS_H

#include "descent.h"
#include "lightning.h"
#include "transprender.h"
#include "particles.h"

#define UNIFY_THREADS	0

typedef struct tRenderThreadInfo {
	tRenderTask	nTask;
	int						nMiddle;
	int						nFaces;
	int						zMax [MAX_THREADS];
	tLightning				*pl;
	int						nLightnings;
	CObject					*objP;
	RenderModel::CModel	*pm;
	tParticleEmitter		*particleEmitters [MAX_THREADS];
	int						nCurTime [MAX_THREADS];
	tThreadInfo				ti [MAX_THREADS];
	} tRenderThreadInfo;

extern tRenderThreadInfo tiRender;

typedef struct tTranspRenderThreadInfo {
	tTranspItemData	itemData [MAX_THREADS];
	tThreadInfo			ti [MAX_THREADS];
	} tTranspRenderThreadInfo;

extern tTranspRenderThreadInfo tiTranspRender;
extern tThreadInfo tiEffects;

int RunRenderThreads (int nTask, int nThreads = -1);
void StartRenderThreads (void);
void EndRenderThreads (void);
void StartTranspRenderThread (void);
void EndTranspRenderThread (void);
void ControlTranspRenderThread (void);
bool WaitForRenderThreads (void);
void StartEffectsThread (void);
void EndEffectsThread (void);
void ControlEffectsThread (void);
bool WaitForEffectsThread (void);
void ControlRenderThreads (void);

int GetNumThreads (void);

//------------------------------------------------------------------------------

inline int RenderThreadsReady (void)
{
return !(gameStates.app.bMultiThreaded && (tiRender.ti [0].bExec || tiRender.ti [1].bExec));
}

//------------------------------------------------------------------------------

static
#ifdef RELEASE
inline
#endif
void ComputeThreadRange (int nId, int nMax, int& nStart, int& nEnd, int nThreads = gameStates.app.nThreads)
{
int nRange = (nMax + nThreads - 1) / nThreads;
nStart = nId * nRange;
nEnd = nStart + nRange;
if (nEnd > nMax)
	nEnd = nMax;
}

//------------------------------------------------------------------------------

#endif // _RENDERTHREADS_H
